// solves 2-D Laplace equation using a relaxation scheme
#define MAX(A,B) (((A) > (B)) ? (A) : (B))
#define ABS(A)   (((A) <  (0)) ? (-A): (A))

#include<omp.h>

#include<stdio.h>
#include<math.h>
#include<float.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>

int main() {
   
  double tol = 0.0;
  double var = DBL_MAX;
  double top = 100.0;
  unsigned n, n2, maxIter;
  int itemsread;
  FILE *fout;

  printf("Enter mesh size, max iterations and tolerance: ");
  itemsread = scanf("%u ,%u ,%lf", &n, &maxIter, &tol);
  
  if (itemsread!=3) {
    fprintf(stderr, "Input error!\n");
    exit(-1);
  }
  
  double time[2] = {0.0, 0.0};
  
  time[0] = omp_get_wtime();
  
  n2 = n+2;
  
  double * T    = calloc(n2*n2, sizeof(double) );
  double * Tnew = calloc(n2*n2, sizeof(double) );
  
  if ( T == NULL || Tnew == NULL ) {
    fprintf(stderr, "Not enough memory!\n");
    exit(EXIT_FAILURE);
  }

  size_t iter = 0;

  // set boundary conditions
#pragma omp parallel
  {
    
#pragma omp for
    for (size_t i=1; i<=n; i++) {
      T[(n+1)*n2 + i    ] = Tnew[(n+1)*n2 + i    ] = i * top / (n+1);  
      T[ i   *n2 + n + 1] = Tnew[i    *n2 + n + 1] = i * top / (n+1);
    }        
    
    while(var > tol && iter <= maxIter) {
      //
      // Synchronize threads at each iteration, to be sure
      // that the vectors have been swapped
      //
#pragma omp barrier
#pragma omp single
      {
	++iter;
	var = 0.0;
      }
#pragma omp for collapse(2) reduction(max:var)
      for (size_t i=1; i<=n; ++i)
	for (size_t j=1; j<=n; ++j) {         
	  Tnew[i*n2+j] = 0.25*( T[(i-1)*n2 + j    ] + T[(i+1)*n2 + j    ] 
				+ T[i    *n2 + (j-1)] + T[i    *n2 + (j+1)] );
	  
	  var = MAX(var, ABS(Tnew[i*n2+j] - T[i*n2+j]));
	}
      
      //
      // Swap pointers
      //
#pragma omp single nowait
      {
	double * Tmp = T;
	T    = Tnew;
	Tnew = Tmp;      
	if (iter%100 == 0)
	  printf("iter: %8u, variation = %12.4lE\n", iter, var);
      }      
    }
  } // omp parallel 
  time[1] = omp_get_wtime();
  
  printf("Elapsed time (s) = %.2lf\n", time[1] - time[0]);
  printf("Mesh size: %u\n", n);
  printf("Stopped at iteration: %u\n", iter);
  printf("Maximum error: %lE\n", var);
  
  // saving results to file
  fout = fopen("results", "w");
  if (fout == NULL) {
    perror("results");
    exit(-1);
  }
  for (size_t i=1; i<=n; ++i)
    for (size_t j=1; j<=n; ++j)
      fprintf(fout, "%8u %8u %18.9lE\n", i, j, T[i*n2+j]);

  fclose(fout);
  free(T);
  free(Tnew);
  return 0;
}
